#ifndef __msn_h__
#define __msn_h__

/*
 * msn.h
 * libmsn
 *
 * Created by Mark Rowe on Mon Apr 17 2004.
 * Refactored by Tiago Salem Herrmann on 08/2007.
 * Copyright (c) 2004 Mark Rowe. All rights reserved.
 * Copyright (c) 2007 Tiago Salem Herrmann. All rights reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <msn/authdata.h>
#include <msn/connection.h>
#include <msn/errorcodes.h>
#include <msn/externals.h>
#include <msn/message.h>
#include <msn/notificationserver.h>
#include <msn/switchboardserver.h>
#include <msn/util.h>
#include <msn/soap.h>

#endif
